/* *************************************************************************
   ************************************************************************* */
   #include  <stdio.h>
   #include  <math.h>
   #include  "cortex.h"

/* =========================================================================
   |                      Structure of Layers                              |
   ========================================================================= */
   struct  input_layer {
           int    Digit_images [CATEGORY][TYPE][ROW][COLUMN];
           float  Digit [ROW][COLUMN]; /* input layer */
   } cell_i;

   struct  hidden_layer {
           float  activity [HUNIT];
           float  threshold [HUNIT];
           float  iweight [HUNIT][ITUPLE];
           int    in_tuple [HUNIT][ITUPLE];
   } cell_h;

   struct  output_layer {
           float  activity [CATEGORY];
           float  threshold [CATEGORY];
           float  weight [CATEGORY][HUNIT];
   } cell_o;


   struct  input_layer        *INPUT;            /* Input layer  */
   struct  hidden_layer       *HIDDEN;           /* Hidden layer */
   struct  output_layer       *OUTPUT;           /* Output layer */

/* =========================================================================
   |                    Functions and Variables                            |
   ========================================================================= */
   void    Read_ntuples();
   void    Read_Digit_images();
   void    Initialize_networks();
   void    Recognize_digit();
   void    Storage_of_outputs();

   FILE    *stream1;                 /* Input file for Digit- images */
   FILE    *stream2;                 /* Input file for hunit's weights */
   FILE    *stream3;                 /* Input file for hunit's thresholds */
   FILE    *stream4;                 /* Input file for ounit's weights */
   FILE    *stream5;                 /* Input file for ounit's thresholds */
   FILE    *stream6;                 /* Input file for random N-tuples */
   FILE    *stream7;                 /* Output file for output results */

main(argc,argv)
int argc;
char *argv[];
{
    register  int   i;               /* Index */
    register  int   j;               /* Index */
    register  int   k;               /* Index */
    register  int   l;               /* Index */

    /* Check the commandline argument */
    if (argc != 3) { printf("Usage : %s inputfile outputfile\n",argv[0]);
                     exit(0); }

    /* Creat and Open files */
      stream1 = fopen ( argv[1], "r" );
      stream2 = fopen ( "hidden.owt", "r" );
      stream3 = fopen ( "hidden.oth", "r" );
      stream4 = fopen ( "output.owt", "r" );
      stream5 = fopen ( "output.oth", "r" );
      stream6 = fopen ( "hidden.tpl","r");
      stream7 = fopen ( argv[2], "w" );

     /* Place the address of structure CELL into the pointer layers */
      INPUT = &cell_i;
      HIDDEN = &cell_h;
      OUTPUT = &cell_o;

    /* Read the random n-tuples for hidden layer */
      Read_ntuples();

    /* Read the Digit images */
      Read_Digit_images();

    /* Initialize the SCHNN */
      Initialize_networks();

    /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
       +                                                                   +
       +                       Recognizing Digits                          +
       +                                                                   +
       +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

      for ( i = 0; i < CATEGORY; i++)
      for ( j = 0; j < TYPE; j++)
      {
          /* Input token */
            for ( k = 0; k < ROW; k++)
            for ( l = 0; l < COLUMN; l++)
            {
                INPUT->Digit[k][l]=(float)INPUT->Digit_images[i][j][k][l];
            }

          /* Recognizing input digit */
            Recognize_digit( i, j );

      }

    /* Close all files and Terminate the process */
      exit(0);

} /* main */

/* =========================================================================
   |                             Functions                                 |
   ========================================================================= */
/* Read the random n-tuples for hidden layer */
void Read_ntuples()
{
    register  int   i;
    register  int   j;

    /* Read n-tuples */
      for ( i = 0; i < HUNIT; i++)
      {
          for ( j = 0; j < ITUPLE; j++)
          {
              fscanf ( stream6, "%4d", &HIDDEN->in_tuple [i][j] );
          }
      }

    return;
} /* Read_ntuples */

/* Read the Digit images */
void Read_Digit_images()
{
    register  int   i;
    register  int   j;
    register  int   k;
    register  int   l;

    /* Read Digit-images */
      for ( i = 0; i < CATEGORY; i++)
      for ( j = 0; j < TYPE; j++)
      for ( k = 0; k < ROW; k++)
      for ( l = 0; l < COLUMN; l++)
      {
          fscanf ( stream1, "%1d", &INPUT->Digit_images [i][j][k][l] );
      }

    return;
} /* Read_Digit_images */

/* Initialize the SCHNN */
void Initialize_networks()
{
    register  int   i;
    register  int   j;

    /* For hidden layer */
      for ( i = 0; i < HUNIT; i++)
      {
          fscanf ( stream3, "%10f", &HIDDEN->threshold [i] );
          for ( j = 0; j < ITUPLE; j++)
          {
              fscanf ( stream2, "%10f", &HIDDEN->iweight [i][j] );
          }
      }

    /* For output layer */
      for ( i = 0; i < CATEGORY; i++)
      {
          fscanf ( stream5, "%10f", &OUTPUT->threshold [i] );
          for ( j = 0; j < HUNIT; j++)
          {
              fscanf ( stream4, "%10f", &OUTPUT->weight [i][j] );
          }
      }

      return;
} /* Initialize_networks */

/* Recognize input digit */
void Recognize_digit( i1, j1)
int  i1, j1;
{
    register  int   i;
    register  int   j;
    register  int   k;
    register  int   l;
    int       tuple = 0;
    int       end;
    int       recogcount = 0;
    float     wsum_hunit = 0.;
    float     wsum_ounit = 0.;
    float     netsum_hunit = 0.;
    float     netsum_ounit = 0.;

    fprintf ( stream7, "Digit = %d  Type = %d\n", i1, j1 );

    do {
       /* Compute the output activities of hidden units */
         for ( i = 0; i < HUNIT; i++)
         {
             for ( j = 0; j < ITUPLE; j++)
             {
                 tuple = HIDDEN->in_tuple [i][j];
                 k = tuple / COLUMN;
                 l = tuple % COLUMN;
                 wsum_hunit += HIDDEN->iweight [i][j] * INPUT->Digit [k][l];
             }

             netsum_hunit = wsum_hunit + HIDDEN->threshold [i];
             HIDDEN->activity [i] = 1. / ( 1. + exp ( -1. * (double)netsum_hunit ) );
             wsum_hunit = 0.0;
         }

       /* Compute the output activities of output units */
         for ( j = 0; j < CATEGORY; j++)
         {
             for ( k = 0; k < HUNIT; k++)
             {
                 wsum_ounit += OUTPUT->weight [j][k] * HIDDEN->activity [k];
             }
             netsum_ounit = wsum_ounit + OUTPUT->threshold [j];
             OUTPUT->activity [j] = 1. / ( 1. + exp ( -1. * netsum_ounit ) );
             wsum_ounit = 0.0;
         }

       /* Write the activation level of output units */
         Storage_of_outputs();

       /* Iteration until some output unit's activity level reachs above 0.9 */
	 for ( j = 0; j < CATEGORY; j++)
         {
             if ( OUTPUT->activity [j] >= 0.9 ) {
                end = TRUE;
		recogcount = 0;
                break; }
             else end = FALSE;
         }

	 /* Try to recognize the digit for three times */
	 ++recogcount;
	 if (recogcount == 3) { recogcount = 0; break; }


    } while ( end == FALSE );

    return;
} /* Recognize_digit */

/* Write the activation level of output units */
void Storage_of_outputs()
{
    register  int   i;

    for ( i = 0; i < CATEGORY; i++)
    {
        fprintf ( stream7, "%8.5f", OUTPUT->activity [i] );
        if ( i == CATEGORY - 1 ) fprintf ( stream7, "\n" );
    }

    return;
} /* Storage_of_outputs */
/* END */









